legistar_host = 'hattiesburg.legistar.com'

DEPARTMENTS_URL = 'https://hattiesburg.legistar.com/Departments.aspx'

PEOPLE_URL = 'https://hattiesburg.legistar.com/People.aspx'


RSS_FEED_URL = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=266278&GUID=a8ed52ea-0528-4c9b-87a7-211c56b0c6d9&Mode=All&Title=City+of+Hattiesburg%2c+MS+-+Calendar+(All)' # noqa

AMENDMENTS_TO_AGENDA_DEPT = 15836

AGENDA_ITEM_ORDER = ['presentation', 'policy', 'routine', 'unknown']

RSS_THIS_MONTH = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=517515&GUID=f9880525-7349-4f51-b375-ac3ded1ccf13&Mode=This+Month&Title=City+of+Hattiesburg,+MS+-+Calendar+(This+Month)' # noqa

RSS_YEAR_2011 = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=517521&GUID=519a4642-ea34-4238-8c5d-68ba5056030f&Mode=2011&Title=City+of+Hattiesburg,+MS+-+Calendar+(2011)' # noqa

RSS_YEAR_2012 = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=517522&GUID=ba87113e-9748-46cc-bed7-09923264420b&Mode=2012&Title=City+of+Hattiesburg,+MS+-+Calendar+(2012)' # noqa

RSS_YEAR_2013 = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=899558&GUID=c71a2185-d86d-484e-b6bc-d7edbfaf07fe&Mode=2013&Title=City+of+Hattiesburg%2c+MS+-+Calendar+%282013%29' # noqa

RSS_YEAR_2014 = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=1304070&GUID=c7f1c7cf-0401-487e-85d8-b25ae27bb0a6&Mode=2014&Title=City+of+Hattiesburg%2c+MS+-+Calendar+%282014%29' # noqa

RSS_YEAR_2015 = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=1800408&GUID=4a65e270-dca1-487b-a50a-775a9e595ecd&Mode=2015&Title=City+of+Hattiesburg%2c+MS+-+Calendar+%282015%29' # noqa

RSS_YEAR_2016 = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=2857417&GUID=0f83b996-05bd-4342-be1c-fc0e267f8dcb&Mode=2016&Title=City+of+Hattiesburg%2c+MS+-+Calendar+%282016%29' # noqa

RSS_YEAR_2017 = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=5663030&GUID=041c7c03-3032-40d2-a64b-2ad48300330f&Mode=2017&Title=City+of+Hattiesburg%2c+MS+-+Calendar+(2017)' # noqa

RSS_YEAR_2018 = 'https://hattiesburg.legistar.com/Feed.ashx?M=Calendar&ID=5663046&GUID=81404e43-eb5a-4a99-808a-8b261de92c2b&Mode=2018&Title=City+of+Hattiesburg%2c+MS+-+Calendar+(2018)' # noqa


RSS_YEARLY_FEEDS = {
    2011: RSS_YEAR_2011,
    2012: RSS_YEAR_2012,
    2013: RSS_YEAR_2013,
    2014: RSS_YEAR_2014,
    2015: RSS_YEAR_2015,
    2016: RSS_YEAR_2016,
    2017: RSS_YEAR_2017,
    2018: RSS_YEAR_2018}



# from sqlalchemy.orm import joinedload
# from sqlalchemy.orm import joinedload_all
# from sqlalchemy.orm import subqueryload
from sqlalchemy.orm.exc import NoResultFound

from .database import Item
from .database import Tag, ItemTag

import transaction


DEFAULT_TAGNAMES = [
    'PlanningCommission',
    'NuncProTunc',
    'Ordinance',
    'CDBG',
    'Entitlement',
    'HOME',
    'ClaimsDocket',
    'MonthlyBudgetReport',
    'FEMA',
    'HomelandSecurity',
    'Appointment',
    ]


def add_tag_names(session):
    for name in DEFAULT_TAGNAMES:
        query = session.query(Tag).filter(Tag.name == name)
        try:
            tag = query.one()
        except NoResultFound:
            # print "adding", name
            tag = Tag()
            tag.name = name
            session.add(tag)
            transaction.commit()


def tag_item(session, item, tagname):
    query = session.query(ItemTag).filter(ItemTag.id == item.id)
    query = query.filter(ItemTag.tag == tagname)
    try:
        item_tag = query.one()
        # msg = "Skipping tagging %s to item %s"
        # print msg % (tagname, item.file_id)
    except NoResultFound:
        item_tag = ItemTag()
        item_tag.id = item.id
        item_tag.tag = tagname
        # msg = "Tagging %s to item %s"
        # print msg % (tagname, item.file_id)
        session.add(item_tag)


def tag_item_for_ordinance(session, item):
    title = item.title
    title = title.lower()
    if 'ordinance' in title:
        tag_item(session, item, 'Ordinance')
        transaction.commit()


def tag_item_for_planning_commission(session, item):
    title = item.title
    if "Hattiesburg Planning Commission" in title:
        tag_item(session, item, 'PlanningCommission')


def tag_item_for_nunc_pro_tunc(session, item):
    title = item.title
    title = title.lower()
    if '(nunc pro tunc)' in title:
        tag_item(session, item, 'NuncProTunc')


def tag_item_for_entitlement(session, item):
    title = item.title
    title = title.lower()
    if '{entitlement}' in title:
        tag_item(session, item, 'Entitlement')


def tag_item_for_cdbg(session, item):
    title = item.title
    if 'CDBG' in title:
        tag_item(session, item, 'CDBG')
    if 'Community Development Block Grant' in title:
        tag_item(session, item, 'CDBG')


def tag_item_for_cdbg_and_entitlement(session, item):
    title = item.title
    if '{Entitlement - CDBG}' in title:
        tag_item(session, item, 'CDBG')
        tag_item(session, item, 'Entitlement')


def tag_item_for_HOME(session, item):
    title = item.title
    if 'HOME Program' in title:
        tag_item(session, item, 'HOME')


def tag_item_for_Claims_Docket(session, item):
    title = item.title
    title = title.lower()
    if 'claims docket' in title:
        tag_item(session, item, 'ClaimsDocket')


def tag_item_for_monthly_budget_report(session, item):
    title = item.title
    title = title.lower()
    if 'monthly budget report' in title:
        tag_item(session, item, 'MonthlyBudgetReport')


def tag_item_for_FEMA(session, item):
    title = item.title
    if ' FEMA ' in title:
        tag_item(session, item, 'FEMA')


def tag_item_for_Homeland_Security(session, item):
    title = item.title
    if 'Homeland Security' in title:
        tag_item(session, item, 'HomelandSecurity')


def tag_item_for_appointment(session, item):
    title = item.title
    title = title.lower()
    if 'appoint' in title:
        tag_item(session, item, 'Appointment')

    pass


TAGGERS = [
    tag_item_for_ordinance,
    tag_item_for_planning_commission,
    tag_item_for_nunc_pro_tunc,
    tag_item_for_entitlement,
    tag_item_for_cdbg,
    tag_item_for_cdbg_and_entitlement,
    tag_item_for_HOME,
    tag_item_for_Claims_Docket,
    tag_item_for_monthly_budget_report,
    tag_item_for_FEMA,
    tag_item_for_appointment,
    ]


def tag_items(session, items):
    for item in items:
        for tagger in TAGGERS:
            tagger(session, item)
        transaction.commit()


def tag_all_items(session):
    query = session.query(Item)
    tag_items(session, query)



from datetime import datetime, date

from sqlalchemy import Column, ForeignKey

# column types
from sqlalchemy import Integer, String, Unicode
from sqlalchemy import Boolean, Date, LargeBinary
from sqlalchemy import PickleType
from sqlalchemy import Enum
from sqlalchemy import DateTime

from sqlalchemy.orm import relationship, backref

from sqlalchemy.ext.declarative import declarative_base

from .legistar import legistar_host

Base = declarative_base()

####################################
# Data Types                      ##
####################################


FileType = Enum('agenda', 'minutes', 'attachment',
                name='lgr_file_type_enum')

AgendaItemType = Enum('presentation', 'policy', 'routine', 'unknown',
                      name='agenda_item_type_enum')

VoteType = Enum('Yea', 'Nay', 'Abstain', 'Absent', 'Present',
                'TEL-No Vote', 'Recuse',
                name='lgr_vote_type_enum')

AgendaItemTypeMap = dict(V='presentation', VI='policy',
                         VII='routine')


CacheType = Enum('action', 'departments', 'item', 'meeting',
                 'people', 'rss-2011', 'rss-2012', 'rss-2013',
                 'rss-this-month',
                 name='lgr_cache_type_enum')


class SerialBase(object):
    def serialize(self):
        data = dict()
        table = self.__table__
        for column in table.columns:
            name = column.name
            try:
                pytype = column.type.python_type
            except NotImplementedError:
                print("NOTIMPLEMENTEDERROR", column.type)
            value = getattr(self, name)
            if pytype is datetime or pytype is date:
                if value is not None:
                    value = value.isoformat()
            data[name] = value
        return data
####################################
#  Tables                         ##
####################################


class MainCache(Base, SerialBase):
    __tablename__ = 'lgr_main_cache'
    id = Column(Integer, primary_key=True)
    name = Column(Unicode(200), unique=True)
    retrieved = Column(DateTime)
    updated = Column(DateTime)
    content = Column(PickleType)


class Department(Base, SerialBase):
    __tablename__ = 'lgr_departments'
    id = Column(Integer, primary_key=True)
    guid = Column(String)
    name = Column(String)

    def __init__(self, id, guid):
        self.id = id
        self.guid = guid
        self.name = None

    def __repr__(self):
        return '<Dept: %d - %s>' % (self.id, self.name)


class Person(Base, SerialBase):
    __tablename__ = 'lgr_people'

    id = Column(Integer, primary_key=True)
    guid = Column(String)
    firstname = Column(String)
    lastname = Column(String)
    website = Column(String)
    photo_link = Column(String)
    notes = Column(String)

    def __init__(self):
        self.id = None
        self.guid = None
        self.firstname = None
        self.lastname = None
        self.website = None
        self.photo_link = None
        self.notes = None

    def __repr__(self):
        msg = '<Person: {} ({} {})>'
        return msg.format(self.id, self.firstname, self.lastname)


class Meeting(Base, SerialBase):
    __tablename__ = 'lgr_meetings'

    id = Column(Integer, primary_key=True)
    guid = Column(String)
    title = Column(String)
    date = Column(Date)
    time = Column(String)
    link = Column(String)
    dept_id = Column(Integer, ForeignKey('lgr_departments.id'))
    agenda_status = Column(String)
    minutes_status = Column(String)
    rss = Column(PickleType)
    updated = Column(DateTime)

    def __init__(self):
        self.id = None
        self.guid = None
        self.title = None
        self.date = None
        self.time = None
        self.link = None
        self.dept_id = None
        self.agenda_link = None
        self.agenda_status = None
        self.minutes_link = None
        self.minutes_status = None
        self.rss = None
        self.updated = None

    def __repr__(self):
        return "<Meeting(%d): '%s'>" % (self.id, self.title)


class Item(Base, SerialBase):
    __tablename__ = 'lgr_items'

    id = Column(Integer, primary_key=True)
    guid = Column(String)
    file_id = Column(String)
    filetype = Column(String)
    name = Column(Unicode)
    title = Column(Unicode)
    status = Column(String)
    passed = Column(Date)
    on_agenda = Column(Date)
    introduced = Column(Date)
    acted_on = Column(Boolean)

    def __init__(self):
        self.id = None
        self.guid = None
        self.file_id = None
        self.filetype = None
        self.name = None
        self.title = None
        self.status = None
        self.passed = None
        self.on_agenda = None
        self.introduced = None
        self.acted_on = False

    def __repr__(self):
        return "<Item: %s, id: %d>" % (self.file_id, self.id)

    def link(self):
        tmpl = 'LegislationDetail.aspx?ID=%d&GUID=%s&Options=&Search='
        return tmpl % (self.id, self.guid)


class MeetingItem(Base, SerialBase):
    __tablename__ = 'lgr_meeting_item'

    meeting_id = Column('meeting_id', Integer,
                        ForeignKey('lgr_meetings.id'),
                        primary_key=True)
    item_id = Column('item_id', Integer,
                     ForeignKey('lgr_items.id'),
                     primary_key=True)

    agenda_num = Column(String)
    type = Column('type', AgendaItemType)
    # order is by type
    order = Column('order', Integer)
    # item order is order of all items
    # this allows unknown council letter to mayor
    # to stay on top of meeting items for sept 18 2012
    item_order = Column('item_order', Integer)
    # I decided to consider keeping track of item
    # versions, and it probably should go in the
    # items table, but for now it is here.
    version = Column('version', Integer)

    def __init__(self, meeting_id, item_id):
        self.meeting_id = meeting_id
        self.item_id = item_id
        self.agenda_num = None
        self.type = None
        self.order = None
        self.item_order = None

    def __repr__(self):
        return "<MeetingItem %d:%d>" % (self.meeting_id, self.item_id)


class Action(Base, SerialBase):
    __tablename__ = 'lgr_actions'

    id = Column(Integer, primary_key=True)
    guid = Column(String)
    file_id = Column(String)
    filetype = Column(String)
    mover_id = Column(Integer, ForeignKey('lgr_people.id'))
    seconder_id = Column(Integer, ForeignKey('lgr_people.id'))
    result = Column(String)
    agenda_note = Column(String)
    minutes_note = Column(String)
    action = Column(String)
    action_text = Column(Unicode)
    # item_id = Column(Integer, ForeignKey('items.id'))

    # related

    def __init__(self):
        self.id = None
        self.guid = None
        self.file_id = None
        self.filetype = None
        self.mover_id = None
        self.seconder_id = None
        self.result = None
        self.agenda_note = None
        self.minutes_note = None
        self.action = None
        self.action_text = None

    def __repr__(self):
        return "<Action: %s, id: %d>" % (self.file_id, self.id)


class ItemAction(Base, SerialBase):
    __tablename__ = 'lgr_item_action'

    item_id = Column('item_id', Integer,
                     ForeignKey('lgr_items.id'),
                     primary_key=True)
    action_id = Column('action_id', Integer,
                       ForeignKey('lgr_actions.id'),
                       primary_key=True)

    def __init__(self, item_id, action_id):
        self.item_id = item_id
        self.action_id = action_id

    def __repr__(self):
        return "<ItemAction %d:%d>" % (self.item_id, self.action_id)


class ActionVote(Base, SerialBase):
    __tablename__ = 'lgr_action_vote'

    action_id = Column('action_id', Integer,
                       ForeignKey('lgr_actions.id'),
                       primary_key=True)
    ward1 = Column('ward1', VoteType)
    ward1_person_id = Column('ward1_person_id', Integer,
                             ForeignKey('lgr_people.id'))

    ward2 = Column('ward2', VoteType)
    ward2_person_id = Column('ward2_person_id', Integer,
                             ForeignKey('lgr_people.id'))
    ward3 = Column('ward3', VoteType)
    ward3_person_id = Column('ward3_person_id', Integer,
                             ForeignKey('lgr_people.id'))
    ward4 = Column('ward4', VoteType)
    ward4_person_id = Column('ward4_person_id', Integer,
                             ForeignKey('lgr_people.id'))
    ward5 = Column('ward5', VoteType)
    ward5_person_id = Column('ward5_person_id', Integer,
                             ForeignKey('lgr_people.id'))


class File(Base, SerialBase):
    __tablename__ = 'lgr_files'

    id = Column(Integer,
                primary_key=True)
    http_info = Column(PickleType)
    content = Column(LargeBinary)
    link = Column(String)

    def __init__(self):
        self.id = None
        self.http_info = None
        self.content = None
        self.link = None

    def __repr__(self):
        return "<File:  id: %d>" % self.id


class Agenda(Base, SerialBase):
    __tablename__ = 'lgr_agendas'

    id = Column(Integer, ForeignKey('lgr_meetings.id'),
                primary_key=True)
    guid = Column(String)
    http_info = Column(PickleType)
    content = Column(LargeBinary)
    link = Column(String)

    def __init__(self):
        self.id = None
        self.guid = None
        self.http_info = None
        self.content = None
        self.link = None

    def __repr__(self):
        return "<Agenda:  id: %d>" % self.id


class Minutes(Base, SerialBase):
    __tablename__ = 'lgr_minutes'

    id = Column(Integer, ForeignKey('lgr_meetings.id'),
                primary_key=True)
    guid = Column(String)
    http_info = Column(PickleType)
    content = Column(LargeBinary)
    link = Column(String)

    def __init__(self):
        self.id = None
        self.guid = None
        self.http_info = None
        self.content = None
        self.link = None

    def __repr__(self):
        return "<Minutes:  id: %d>" % self.id


class Attachment(Base, SerialBase):
    __tablename__ = 'lgr_attachments'

    id = Column(Integer, primary_key=True)
    guid = Column(String)
    name = Column(String)
    http_info = Column(PickleType)
    content = Column(LargeBinary)
    link = Column(String)
    item_id = Column(Integer, ForeignKey('lgr_items.id'))

    def __init__(self):
        self.id = None
        self.guid = None
        self.name = None
        self.http_info = None
        self.content = None
        self.link = None
        self.item_id = None

    def __repr__(self):
        return "<Attachment:  id: %d>" % self.id

    def get_link(self):
        return 'https://%s/%s' % (legistar_host, self.link)


class Tag(Base, SerialBase):
    __tablename__ = 'lgr_tagnames'

    name = Column(String, primary_key=True)

    def __init__(self):
        self.name = None

    def __repr__(self):
        return "<Tag: %s>" % self.name


class ItemTag(Base, SerialBase):
    __tablename__ = 'lgr_item_tags'

    id = Column(Integer, ForeignKey('lgr_items.id'),
                primary_key=True)
    tag = Column(String, ForeignKey('lgr_tagnames.name'),
                 primary_key=True)

    def __init__(self):
        self.id = None
        self.tag = None

    def __repr__(self):
        return "<ItemTag: %s:%s>" % (self.id, self.tag)


# ItemTag relationships
Tag.items = relationship(Item, backref='tags',
                         order_by=Item.id,
                         secondary='lgr_item_tags')


#######################################################
#######################################################

Department.meetings = relationship(Meeting, order_by=Meeting.date)

# Meeting relationships
meeting_backref = backref('meeting', uselist=False)

Meeting.dept = relationship(Department)


Meeting.items = relationship(Item, backref='meetings',
                             secondary='lgr_meeting_item')

Meeting.meeting_items = relationship(MeetingItem,
                                     order_by=MeetingItem.item_order)

Meeting.agenda = relationship(Agenda, uselist=False,
                              backref=meeting_backref)
Meeting.minutes = relationship(Minutes, uselist=False,
                               backref=meeting_backref)


# MeetingItem relationships
MeetingItem.meeting = relationship(Meeting)
MeetingItem.item = relationship(Item)


# Item relationships
Item.actions = relationship(Action, backref='items',
                            order_by=Action.id,
                            secondary='lgr_item_action')

Item.attachments = relationship(Attachment, backref='item',
                                order_by=Attachment.id)


# Action relationships
Action.item = relationship(Item, backref='items',
                           secondary='lgr_item_action')

Action.mover = relationship(Person, primaryjoin='Action.mover_id == Person.id')
Action.seconder = relationship(Person,
                               primaryjoin='Action.seconder_id == Person.id')



import random
import time
import datetime
import urllib.parse
import urllib.request
import urllib.error
import urllib.parse
import http.client
import http.client.IncompleteRead

from .legistar import legistar_host


class FileExistsError(Exception):
    pass


class BadDownloadError(Exception):
    pass


class InvalidDateFormat(Exception):
    pass


# Tidy function inspired by:
# http://www.toao.net/48-replacing-smart-quotes-and-em-dashes-in-mysql
def tidy_needless_utf_punctuation(data):
    data = data.replace('\xe2\x80\x98', "'")
    data = data.replace('\xe2\x80\x99', "'")
    data = data.replace('\xe2\x80\x9c', '"')
    data = data.replace('\xe2\x80\x9d', '"')
    data = data.replace('\xe2\x80\x93', "-")
    data = data.replace('\xe2\x80\x94', "--")
    data = data.replace('\xe2\x80\xa6', "...")
    return data


def convert_range_to_datetime(start, end):
    "start and end are timestamps"
    start = datetime.datetime.fromtimestamp(float(start))
    end = datetime.datetime.fromtimestamp(float(end))
    return start, end


def random_wait(minimum=5, maximum=15, msg=''):
    # seconds = random.randint(minimum, maximum)
    seconds = random.random() * 5 + 1
    if msg:
        template_data = dict(seconds=seconds)
        msg = msg % template_data
        print(msg)
    time.sleep(seconds)


def parse_date_string(datestring):
    dlist = datestring.split('/')
    if len(dlist) != 3:
        raise InvalidDateFormat("Invalid date string: %s" % datestring)
    dlist = list(map(int, dlist))
    proper = dlist[2], dlist[0], dlist[1]
    return proper


def make_true_date(datestring):
    proper = parse_date_string(datestring)
    return datetime.date(*proper)


def parse_legistar_cgi_query(link):
    if type(link) is bytes:
        link = link.decode()
    uparse = urllib.parse.urlparse(link)
    query = uparse.query
    item_split = query.split('&')
    items = [item.split('=') for item in item_split]
    parsed = dict(items)
    return parsed


def legistar_id_guid(link):
    if type(link) is bytes:
        link = link.decode()
    plink = parse_legistar_cgi_query(link)
    id = int(plink['ID'])
    guid = plink['GUID']
    return id, guid


def onclick_link(attribute):
    return attribute.split("'")[1]


def rss_entry_updated(entry):
    uparsed = entry.updated_parsed
    updated = datetime.datetime(*uparsed[:6])
    return updated


def _view_url(id, guid, type):
    """Returns the url for for the agenda(A)/minutes(M) from a meeting identified
    by id, guid"""
    host = legistar_host
    url_template = 'https://%s/View.ashx?M=%s&ID=%d&GUID=%s'
    url = url_template % (host, type, id, guid)
    return url


def agenda_url(id, guid):
    """Returns the url for for the agenda from a meeting identified
    by id, guid"""
    return _view_url(id, guid, 'A')


def minutes_url(id, guid):
    """Returns the url for for the agenda from a meeting identified
    by id, guid"""
    return _view_url(id, guid, 'M')


#################################
#################################
#################################
#################################
# Handle download stuff below
#################################
#################################
#################################
#################################
#################################
def handle_link(uri):
    data = dict()
    attachment_marker = 'attachment;'
    filename_marker = 'filename='
    disposition_key = 'content-disposition'
    length_key = 'content-length'
    f = urllib.request.urlopen(uri)
    info = f.info()
    data['info'] = info
    data['fileobj'] = f
    # presume no attachment first
    data['attachment'] = False
    if disposition_key in info:
        disposition = info[disposition_key]
        if disposition.startswith(attachment_marker):
            data['attachment'] = True
            fragment = disposition.split(attachment_marker)[1]
            filename = fragment.split(filename_marker)[1]
            length = int(info[length_key])
            data['filename'] = filename
            data['length'] = length
    return data


# http://stackoverflow.com/a/14206036/1869821
# http://bobrochel.blogspot.co.nz/2010/11/bad-servers-chunked-encoding-and.html
def patch_http_response_read(func):
    def inner(*args):
        try:
            return func(*args)
        except http.client.IncompleteRead as e:
            return e.partial
    return inner


http.client.HTTPResponse.read = patch_http_response_read(http.client.HTTPResponse.read) # noqa


def get_rss_feed(url):
    response = urllib.request.urlopen(url)
    return response.read(), response.info()



from sqlalchemy import or_
from sqlalchemy.exc import IntegrityError

from .database import Meeting, Item, Action
from .database import Person, Department

from .manager import ModelManager, delete_all

from .util import legistar_id_guid


def check_dupe_item(session, item):
    Dupe = True
    # FIXME straighten this out
    if 'id' not in item:
        print("Bad item", item)
        import pdb ; pdb.set_trace() # noqa
        return Dupe
    dbitem = session.query(Item).get(item['id'])
    if dbitem is None:
        return False
    filtered_keys = ['attachments', 'action_details']
    keys = [k for k in list(item.keys()) if k not in filtered_keys]
    for key in keys:
        if getattr(dbitem, key) != item[key]:
            Dupe = False
    return Dupe


class DatabaseManager(object):
    def __init__(self, session, collector):
        self.session = session
        self.collector = collector
        self.manager = ModelManager(self.session)
        self.meeting = None
        self.parsed = None

    @property
    def collecter(self):
        # FIXME collecter should be collector
        import warnings
        warnings.warn("Please use .collector instead.", stacklevel=2)
        return self.collector

    def add_collected_people(self, people):
        return self.manager.add_collected_people(people)

    def add_people(self):
        people = self.session.query(Person).all()
        if not len(people):
            print("adding people........")
            people = self.collector.collect('people')
            self.add_collected_people(people)
            self.session.commit()

    def add_collected_depts(self, depts):
        return self.manager.add_collected_depts(depts)

    def add_departments(self):
        depts = self.session.query(Department).all()
        if not len(depts):
            print("adding departments.....")
            depts = self.collector.collect('depts')
            self.add_collected_depts(depts)
            self.session.commit()

    def add_rss_meetings(self, url, rss=None):
        if rss is None:
            rss = self.manager.get_rss(url)
        for entry in rss.entries:
            id, guid = legistar_id_guid(entry.link)
            meeting = self.session.query(Meeting).get(id)
            if meeting is None:
                print("adding meeting {} from rss".format(id))
                try:
                    self.manager.add_meeting_from_rss(entry)
                    self.session.commit()
                except IntegrityError:
                    self.session.rollback()
            else:
                print("Meeting {} already present.".format(id))

    def set_meeting(self, meeting):
        self.meeting = meeting
        self.parsed = self.collector.collect('meeting', link=self.meeting.link)

    def add_meetings(self):
        "This adds full meetings which have been added to meetings table by rss." # noqa
        meetings = self.session.query(Meeting).all()
        for meeting in meetings:
            self.set_meeting(meeting)
            self.add_meeting(meeting)
        self.session.commit()

    def add_meeting(self, meeting):
        for item in self.parsed['items']:
            self.add_meeting_item(item)
        self.manager._merge_collected_meeting_items(meeting, self.parsed)
        self.manager._merge_collected_meeting(meeting, self.parsed)
        self.session.commit()

    def add_pickled_meeting(self, meeting):
        meeting_id = meeting['info']['id']
        for item in meeting['items']:
            self.add_collected_item(item)
        self.manager._merge_pickled_meeting_items(meeting_id, meeting['items'])

    def merge_nonfinal(self):
        filtered = or_(Meeting.minutes_status == None, # noqa
                       Meeting.minutes_status != 'Final')
        qmeetings = self.session.query(Meeting).filter(filtered)
        meetings = qmeetings.all()
        for meeting in meetings:
            collected = self.collector.collect('meeting', link=meeting.link)
            self.manager._merge_collected_meeting(meeting, collected)
            print("Merging meeting %d" % meeting.id)
            self.session.commit()

    def convert_binary_item(self, item):
        if b'id' in item:
            sitem = {}
            for key in item:
                sitem[key.decode()] = item[key]
            item = sitem
        return item

    def add_collected_actions(self, item_id, actions):
        for action in actions:
            dbaction = self.session.query(Action).get(action['id'])
            if dbaction is None:
                print("adding action %d to database." % action['id'])
                self.manager.add_collected_action(item_id, action)
        self.session.commit()

    def add_collected_item(self, item):
        item = self.convert_binary_item(item)
        if not check_dupe_item(self.session, item):
            print("adding item %d to database." % item['id'])
            self.manager._add_collected_legislation_item(item)
        self.add_collected_actions(item['id'], item['actions'])
        self.session.commit()

    def add_meeting_item(self, parsed_item):
        item = self.collector.collect('item', link=parsed_item['item_page'])
        item = self.convert_binary_item(item)
        if not check_dupe_item(self.session, item):
            print("adding item %d to database." % item['id'])
            self.manager._add_collected_legislation_item(item)
        if 'action_details' not in item:
            action_details = []
        else:
            action_details = item['action_details']
        self.add_item_actions(item['id'], action_details)

    def add_item_actions(self, item_id, action_details):
        for link in action_details:
            action = self.collector.collect('action', link=link)
            dbaction = self.session.query(Action).get(action['id'])
            if dbaction is None:
                print("adding action %d to database." % action['id'])
                self.manager.add_collected_action(item_id, action)
        self.session.commit()

    def delete_all(self):
        delete_all(self.session)



from datetime import datetime
import transaction

from sqlalchemy.orm.exc import NoResultFound
# from sqlalchemy import desc

# from .legistar import legistar_host
from .util import legistar_id_guid
from .util import make_true_date
from .util import convert_range_to_datetime

from .database import Department, Person
from .database import Meeting, Item, MeetingItem
from .database import ItemAction, Action, ActionVote
from .database import Attachment
from .database import AgendaItemTypeMap

from .collector.main import MainCollector
from .collector.rss import RssCollector

timeformat = '%I:%M %p'

drop_models = [Attachment, ActionVote, Action, ItemAction,
               MeetingItem, Item, Meeting, Department, Person]


def delete_all(session):
    transaction.begin()
    for model in drop_models:
        q = session.query(model)
        q.delete()
    transaction.commit()


def convert_agenda_number(agenda_number):
    while agenda_number.startswith('.'):
        agenda_number = agenda_number[1:]
    delimiter = '.-'
    for delimiter in ['. -', '.-', '. - ', ' - ', '-']:
        if delimiter in agenda_number:
            break
    if delimiter in agenda_number:
        itemtype, order = agenda_number.split(delimiter)
        if itemtype.startswith('+'):
            itemtype = itemtype[1:]
        itemtype = AgendaItemTypeMap[itemtype]
        order = order.strip()
        while order.endswith('.'):
            order = order[:-1]
        order = int(order)
    else:
        itemtype = 'unknown'
        while agenda_number.endswith('.'):
            agenda_number = agenda_number[:-1]
        if agenda_number:
            order = int(agenda_number)
        else:
            order = None
    return itemtype, order


class ModelManager(object):
    def __init__(self, session):
        self.session = session

    def set_session(self, session):
        self.session = session

    def collector(self):
        return MainCollector()

    def _range_filter(self, query, start, end):
        query = query.filter(Meeting.date >= start)
        query = query.filter(Meeting.date <= end)
        return query

    def get_ranged_meetings(self, start, end, timestamps=False):
        if timestamps:
            start, end = convert_range_to_datetime(start, end)
        q = self.session.query(Meeting)
        q = self._range_filter(q, start, end)
        return q.all()

    # entry is rss entry
    def add_meeting_from_rss(self, entry):
        transaction.begin()
        meeting = Meeting()
        meeting.title = entry.title
        meeting.link = entry.link
        meeting.rss = entry
        meeting.id, meeting.guid = legistar_id_guid(entry.link)
        meeting.updated = datetime.now()
        self.session.add(meeting)
        self.session.flush()
        transaction.commit()

    # retrieve basic meeting info from
    # meeting details page on legistar.
    # the link is from the rss entry
    def remote_meeting(self, link):
        collector = self.collector()
        collector.set_url(link)
        collector.collect('meeting')
        return collector.meeting

    def remote_meeting_info(self, link):
        meeting = self.remote_meeting(link)
        info = meeting['info']
        return info

    def remote_meeting_items(self, link):
        meeting = self.remote_meeting(link)
        items = meeting['items']
        return items

    # link is relative from legistar prefix
    def _remote_legislation_item(self, link):
        print("using link", link)
        collector = self.collector()
        url = collector.url_prefix + link
        collector.set_url(url)
        collector.collect('item')
        return collector.item

    # link is relative url to item page
    def remote_legislation_item(self, link):
        item = self._remote_legislation_item(link)
        # add id, guid to item
        item['id'], item['guid'] = legistar_id_guid(link)
        for key in ['introduced', 'on_agenda', 'passed']:
            if key in item and item[key]:
                item[key] = make_true_date(item[key])
        key = 'action_details'
        if len(item[key]):
            item['acted_on'] = True
        else:
            item['acted_on'] = False
        return item

    # link is full url to meeting
    def remote_legislation_items(self, link):
        meeting_items = self.remote_meeting_items(link)
        print("Meeting has %d items" % len(meeting_items))
        leg_items = []
        for item in meeting_items:
            item_page = item['item_page']
            leg_item = self.remote_legislation_item(item_page)
            leg_items.append(leg_item)
        return leg_items

    def merge_remote_meeting_items(self, meeting_id):
        pass

    def merge_remote_legislation_items(self, meeting_id):
        pass

    # meeting is db object, collected is from collector
    def _merge_collected_meeting(self, meeting, collected):
        transaction.begin()
        info = collected['info']
        for key in ['id', 'guid', 'time', 'link',
                    'dept_id', 'agenda_status', 'minutes_status']:
            value = info[key]
            setattr(meeting, key, value)
        setattr(meeting, 'date', make_true_date(info['date']))
        meeting.updated = datetime.now()
        self.session.merge(meeting)
        self.session.flush()
        transaction.commit()

    def _merge_collected_meeting_items(self, meeting, collected):
        meeting_id = meeting.id
        return self._merge_pickled_meeting_items(meeting_id, collected)

    def _merge_pickled_meeting_items(self, meeting_id, collected):
        transaction.begin()
        items = collected['items']
        item_count = 0
        for item in items:
            item_count += 1
            item_id, guid = legistar_id_guid(item['item_page'])
            query = self.session.query(MeetingItem)
            query = query.filter_by(meeting_id=meeting_id)
            query = query.filter_by(item_id=item_id)
            try:
                dbitem = query.one()
            except NoResultFound:
                dbitem = MeetingItem(meeting_id, item_id)
            agenda_num = item['agenda_num']
            ##########################################
            # # Work around       ####################
            # # irregular entries ####################
            ##########################################
            if agenda_num == '2011-0229':
                agenda_num = None
            if agenda_num == '1.':
                agenda_num = '1'
            ##########################################
            #                     ####################
            #                     ####################
            ##########################################
            # first agenda item is missing from meeting details
            if meeting_id == 302621:
                agenda_num = '2'
            if agenda_num is not None:
                dbitem.agenda_num = agenda_num
                dbitem.type, dbitem.order = convert_agenda_number(agenda_num)
            dbitem.item_order = item_count
            dbitem.version = int(item['version'])
            self.session.merge(dbitem)
            self.session.flush()
        transaction.commit()

    def merge_meeting_from_legistar(self, id):
        collector = MainCollector()
        meeting = self.session.query(Meeting).filter_by(id=id).one()
        collector.set_url(meeting.link)
        collector.collect('meeting')
        self._merge_collected_meeting(meeting, collector.meeting)

    def add_departments(self):
        collector = MainCollector()
        collector.collect('dept')
        self.add_collected_depts(collector.result)

    def add_collected_depts(self, depts):
        transaction.begin()
        for dept_info in depts:
            id, guid, name = dept_info
            dept = Department(id, guid)
            dept.name = name
            self.session.add(dept)
        self.session.flush()
        transaction.commit()

    def update_departments(self):
        collector = MainCollector()
        collector.collect('dept')
        transaction.begin()
        for dept_info in collector.result:
            id, guid, name = dept_info
            dept = self.session.query(Department).get(id)
            if dept is None:
                dept = Department(id, guid)
                dept.name = name
                self.session.add(dept)
                continue
            dept.id = id
            dept.guid = guid
            dept.name = name
            self.session.add(dept)
        self.session.flush()
        transaction.commit()

    def add_people(self):
        collector = MainCollector()
        collector.collect('people')
        self.add_collected_people(collector.result)

    def add_collected_people(self, people):
        transaction.begin()
        for pinfo in people:
            person = Person()
            for key in pinfo:
                setattr(person, key, pinfo[key])
            self.session.add(person)
        self.session.flush()
        transaction.commit()

    def get_rss(self, url):
        collector = RssCollector()
        collector.get_rss(url)
        return collector

    def add_collected_action(self, item_id, action):
        dbaction = Action()
        action['filetype'] = action['ftype']
        main_keys = ['id', 'guid']
        if not action['roll_call']:
            main_keys += ['action', 'action_text', 'result', 'filetype']
        for key in main_keys:
            setattr(dbaction, key, action[key])
        if not action['roll_call']:
            for key in ['agenda_note', 'minutes_note']:
                value = action[key].strip()
                if not value:
                    value = None
                setattr(dbaction, key, value)
            for key in ['mover', 'seconder']:
                name, link = action[key]
                if not name:
                    continue
                id, guid = legistar_id_guid(link)
                att = '%s_id' % key
                setattr(dbaction, att, id)
            file_id, link = action['file_id']
            dbaction.file_id = file_id
        else:
            dbaction.action = 'Roll Call'
        self.session.add(dbaction)
        # flush here so the action can be referred by
        # foreign keys in 'item_action' and 'action_vote'
        self.session.flush()
        # make item_action object
        item_action = ItemAction(item_id, dbaction.id)
        self.session.add(item_action)
        # handle votes
        ward_num = 1
        vote_attributes = dict(action_id=dbaction.id)
        for name, link, vote in action['votes']:
            person_id, ignore = legistar_id_guid(link)
            id_key = 'ward{}_person_id'.format(ward_num)
            vote_attributes[id_key] = person_id
            vote_attributes['ward{}'.format(ward_num)] = vote
            ward_num += 1

        avote = ActionVote()
        for attribute, value in vote_attributes.items():
            setattr(avote, attribute, value)
        self.session.add(avote)

    # add the item before the actions
    def _add_collected_legislation_item(self, item):
        transaction.begin()
        dbitem = Item()
        for key in item:
            if key == 'attachments':
                continue
            value = item[key]
            setattr(dbitem, key, value)
        self.session.add(dbitem)
        if item['attachments'] is not None:
            for name, link in item['attachments']:
                id, guid = legistar_id_guid(link)
                dbobj = self.session.query(Attachment).get(id)
                if dbobj is None:
                    attachment = Attachment()
                    attachment.id, attachment.guid = id, guid
                    attachment.name = name
                    attachment.link = link
                    attachment.item_id = dbitem.id
                    self.session.add(attachment)
                else:
                    msg = 'Duplicate attachment %d' % id
                    raise RuntimeError(msg)
        transaction.commit()

    # here item is an item collected from
    # legistar
    def add_new_legislation_item(self, item):
        self._add_collected_legislation_item(item)
        for link in item['action_details']:
            collector = MainCollector()
            url = collector.url_prefix + link
            collector.set_url(url)
            collector.collect('action')
            self.add_collected_action(item['id'], collector.action)
        self.session.flush()
        transaction.commit()
